
(* Interface to the driver structure. - Tom 7 *)
signature DRIVER =
sig

  datatype optimization =
    Iltoil of Il.exp -> Il.exp
  | Liltolil of Lil.exp -> Lil.exp
  | Iltolil of Il.exp -> Lil.exp
  | Liltoil of Lil.exp -> Il.exp
(* ? perry's thing *)

  (* call this function in your optimization's module,
     which will register the optimization to be run
     unless disabled is true. 

     if trusted is set, we don't run the validator
     after the optimization. Don't set this unless
     you're *really* sure.

     if blowup is set, then we allow the expression to
     potentially increase in size from the optimization.
     (some optimizations may be run simply to enable others.)
     
     init score initializes the 'lastcall' field; opts with higher
     values will be run before those with lower values.

     scorefn asks you to predict how well the next call will
     do based on how well this one will do.
     (typically fn x=>x or fn x=>0)
  *)
  val register_opt : {optname   : string,
                      disabled  : bool,
                      trusted   : bool,
                      initscore : int option,
                      blowup    : bool,
                      scorefn   : int -> int,
                      optfun    : optimization} -> unit

  (* optimize the given expression, with a guess
     of n seconds remaining. Then print the expression. *)
  val optimize : {timer : Timer.real_timer,
                  seconds : int,
                  exp : Lil.exp} -> unit

  val drivers : unit -> (string * optimization) list

end
